/*

 path.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "path.h"
#include "partition.h"
#include <time.h>
#include <sys/stat.h>
#include <string.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include "libtree.h"
#include "shorten.h"

extern int relaxfiletypes;
extern FILE *lf;
extern void errtochannel15(int);

static inline unsigned char topetscii(char c)
{
    if (c >= 'A' && c <= 'Z') return c + 128;
    if (c >= 'a' && c <= 'z') return c - 32;
    if (c == '_') return 0xA4;
    return c;
}

char c64toascii(unsigned char c)
{
    if (c >= 0x41 && c <= 0x5A) return c + 32; // 'a'
    if (c >= 0x61 && c <= 0x7A) return c - 32; // 'A'
    if (c == 0xA4) return '_';
    if (c >= 0xC1 && c <= 0xDA) return c - 128;// 'A'
    return c;
}

struct filename_s {
    char name[17];
    char type[4];
    struct avltree_node node;
};

static struct avltree filenames;
static struct filename_s *lastfn;

static int filename_compare(const struct avltree_node *aa, const struct avltree_node *bb)
{
    const struct filename_s *a = cavltree_container_of(aa, struct filename_s, node);
    const struct filename_s *b = cavltree_container_of(bb, struct filename_s, node);
    int i = strcmp((char *)&a->name, (char *)&b->name);
    if (i) return i;
    return strcmp((char *)&a->type, (char *)&b->type);
}

static void filename_free(struct avltree_node *aa)
{
    struct filename_s *a = avltree_container_of(aa, struct filename_s, node);
    free(a);
}

void initdirentry(int i) {
    if (i) {
        avltree_destroy(&filenames, filename_free);
        free(lastfn);
        lastfn = NULL;
    }
    else avltree_init(&filenames);
}

int createdirentry(const char *filename, const char *lname2, DIRECTORY_ENTRY_T *kesz)
{
    struct tm ido2;
    struct avltree_node *b;
    struct stat buf;

    if (lstat(lname2, &buf)) return 1;

    ido2 = *localtime(&buf.st_mtime);
    ido2.tm_mon++;
    kesz->time[0] = (ido2.tm_sec % 60) | ((ido2.tm_mon << 4) & 0xc0);
    kesz->time[1] = (ido2.tm_min % 60) | (ido2.tm_mon << 6);
    kesz->time[2] = (((ido2.tm_year - 80) % 100) & 63) | ((ido2.tm_hour << 3) & 0xc0);
    kesz->time[3] = (ido2.tm_mday & 31) | (ido2.tm_hour << 5);
    memset(kesz->pointer, 0, 4);
    memset(kesz->name, 0, 16);
    if (relaxfiletypes) {
        memcpy(kesz->filetype, "PRG", 3);
    } else {
        memset(kesz->filetype, 0, 3);
    }
    kesz->attrib = A_DELETEABLE;
    if (buf.st_mode & S_IXUSR) {
        kesz->attrib |= A_EXECUTEABLE;
    }
    if (buf.st_mode & S_IRUSR) {
        kesz->attrib |= A_READABLE;
    }
    if (buf.st_mode & S_IWUSR) {
        kesz->attrib |= A_WRITEABLE;
    }
    kesz->filesize[0] = (unsigned int)buf.st_size;
    kesz->filesize[1] = (unsigned int)buf.st_size >> 8;
    kesz->filesize[2] = (unsigned int)buf.st_size >> 16;
    kesz->filesize[3] = (unsigned int)buf.st_size >> 24;
    if (strcmp(filename, ".") && strcmp(filename, "..")) {
        const char *tmp = filename;
        size_t i, j, k, l;
        if (tmp[0] == '.') {
            kesz->pointer[0] |= 0x80;
        }

        j = 0; l = k = strlen(tmp);
        for (i = 0; i < k; i++) {
            switch (tmp[i]) {
            case '.':
                if (S_ISDIR(buf.st_mode) || relaxfiletypes) break;
                /* fall through */
            case ',':
                memset(kesz->filetype, 0, 4);
                l = j = i;
                /* fall through */
            default:
                break;
            }
        }
        if (l > 16) {
            tmp = shorten(filename, l);
            if (!tmp) return 1;
            l = 16;
        } 

        for (i = 0; i < l; i++) {
            char c = tmp[i];
            if (c == ':' || c == '=' || c == '*' || c == '?' || c == ',') {
                c = '_';
            }
            kesz->name[i] = topetscii(c);
        }
        if (tmp != filename) free((char *)tmp);

        if (j) {
            for (i = j + 1; i < k; i++) {
                char c = filename[i];
                if ((i - j) > 3 || c == 0) {
                    break;
                }
                if (c == ':' || c == '=' || c == '*' || c == '?' || c == ',' || c == '<' || c == ' ') {
                    c = '_';
                }
                kesz->filetype[i - j - 1] = topetscii(c);
            }
        }
    } else {
        if (!filename[1]) {
            kesz->pointer[0] |= 0x80;
        }
        strncpy((char *)kesz->name, filename, 16);
    }

    if (S_ISDIR(buf.st_mode)) {
        memcpy(kesz->filetype, "DIR", 3);
        kesz->attrib |= A_DIR | A_CLOSED;
        memset(kesz->filesize, 0, 4);
    } else if (S_ISREG(buf.st_mode)) {
        kesz->attrib |= A_NORMAL | A_CLOSED;
    } else if (S_ISLNK(buf.st_mode)) {
        memcpy(kesz->filetype, "LNK", 3);
        kesz->attrib |= A_LNK | A_CLOSED;
    } else {
        memcpy(kesz->filetype, "DEL", 3);
        kesz->attrib = A_DEL | A_CLOSED;
        memset(kesz->filesize, 0, 4);
    }

    if (!lastfn) lastfn = (struct filename_s *)malloc(sizeof(struct filename_s));
    if (!lastfn) return 1;

    strncpy(lastfn->name, (char *)kesz->name, 16);
    lastfn->name[16] = 0;
    strncpy(lastfn->type, (char *)kesz->filetype, 3);
    lastfn->type[3] = 0;
    b = avltree_insert(&lastfn->node, &filenames, filename_compare);
    if (b) return 1;
    lastfn = NULL;
    return 0;
}

void convertfilename(const c64c_t *s, char typek, c64c_t *name, c64c_t *type, unsigned char *mode)
{
    int b, i, j;
    char type2[4], mode2;
    int comma;

    i = comma = mode2 = 0; j = -1;

    for (b = 0; s[b] && s[b] != '/'; b++) {
        if (s[b] == ':') {
            i = comma = mode2 = 0; j = -1; continue;
        }

        switch (comma) {
        case 0:
            if (s[b] == typek) {
                comma++; j = 0;
                break;
            }
            if (i < 16) {
                name[i++] = s[b];
            }
            break;
        case 1:
            if (s[b] == ',') {
                comma++;
                break;
            }
            if (j < 3) {
                type2[j++] = s[b];
            }
            break;
        case 2:
            mode2 = s[b];
            comma++; break;
        default:
            break;
        }
    }
    name[i] = 0;
    if (j >= 0 && type) {
        if (j == 1) {
            switch (type2[0]) {
            case 'S': strcpy(type2, "SEQ"); j = 3;break;
            case 'P': strcpy(type2, "PRG"); j = 3;break;
            case 'U': strcpy(type2, "USR"); j = 3;break;
            case 'B': strcpy(type2, "DIR"); j = 3;break;
            case 'J': strcpy(type2, "LNK"); j = 3;break;
            }
        }
        memcpy(type, type2, j);
        type[j] = 0;
    }
    if (mode2 && mode) {
        *mode = mode2;
    }
}

int matchname(c64c_t *a, c64c_t *b)
{
    while (*a == *b || (*b == '?' && *a != 0)) {
        if (*a == 0) {
            return 1;
        }
        a++; b++;
    }

    if (*b == '*') {
        b++;
        if (*b == 0) {
            return 1;
        }
        while (*a != 0) {
            if (matchname(a, b)) {
                return 1;
            }
            a++;
        }
    }

    return 0;
}

int resolv_path(const c64c_t *s, char *outpath, const c64c_t **outname, unsigned char *outpart)
{
    c64c_t name[17];
    struct dirent *ep;
    DIR *dp;
    int found;
    int fel = 0, fel2 = 0;
    unsigned char partition = 0;

    if (strchr((char *)s, ':')) {
        while ((s[fel] ^ 0x30) < 10) partition = partition * 10 + (s[fel++] ^ 0x30);
        while (s[fel] && s[fel] != '/' && s[fel] != ':') {
            fel++;
        }
        fel++;
    }

    if (s[fel] == '/') {
        outpath[0] = 0;
        found = 1;
    } else {
        strcpy(outpath, partition_get_path(partition));
        found = 0;
    }

    while (1) {
        if (s[fel] == '/' || s[fel] == ':') {
            fel++; continue;
        }

        if (!s[fel]) {
            *outname = &s[fel];
            if (outpart) *outpart = partition;
            return found;
        }
        if (!strcmp((char *)&s[fel], "./")) {
            fel += 2; continue;
        }
        if (!strcmp((char *)&s[fel], "../")) {
            char *o = strrchr(outpath, '/');
            if (o) {
                *o = 0;
            } else {
                outpath[0] = 0;
            }
            fel += 3; continue;
        }

        fel2 = fel;
        while (s[fel2] && s[fel2] != '/' && s[fel2] != ':') {
            fel2++;
        }
        if (!s[fel2]) {
            *outname = &s[fel];
            if (outpart) *outpart = partition;
            return 1;
        }

        dp = opendir(outpath[0] ? outpath : ".");
        if (!dp) {
            errtochannel15(1); fprintf(lf, "Couldn't open the directory \"%s\".\n", outpath); return 0;
        }
        convertfilename(s + fel, '/', name, NULL, NULL);

        initdirentry(0);
        while ((ep = readdir(dp)))
        {
            DIRECTORY_ENTRY_T dirent;
            char lname2[2020];

            strcpy(lname2, outpath);
            if (outpath[0]) {
                strncat(lname2, "/", 999);
            }
            strncat(lname2, ep->d_name, 999);

            if (createdirentry(ep->d_name, lname2, &dirent)) continue;

            if ((dirent.attrib & (A_DIR | A_CLOSED)) != (A_CLOSED | A_DIR)) {
                continue;
            }

            dirent.filesize[0] = 0;
            if (!matchname(dirent.name, name)) {
                continue;
            }

            found = 1;
            break;
        }
        initdirentry(1);

        if (!ep || (strlen(ep->d_name) + strlen(outpath) + 2) > 1000) {
            closedir(dp);
            return 0;
        }

        if (outpath[0]) {
            strcat(outpath, "/");
        }
        strcat(outpath, ep->d_name);
        closedir(dp);
        while (s[fel] && s[fel] != '/' && s[fel] != ':') {
            fel++;
        }
    }
}
